
export const enum KeyEventType {
	IfCaught,
	OnlyOnce,
	AtLeastOnce,
}
type QueueItem<T> = {
	outputValue: T;
	eventType: KeyEventType;
	polled: boolean;
};

export class Keyboard<T> {
	private keysPressed: Set<T>;
	private keyQueue: Map<T, QueueItem<T>>;
	private keyMap: Map<string, QueueItem<T>>;
	constructor() {
		this.keyQueue = new Map();
		this.keyMap = new Map();
		this.keysPressed = new Set();
		document.addEventListener("keydown", (ke) => {
			this.keyDown(ke.key);
		});

		document.addEventListener("keyup", (ke) => {
			this.keyUp(ke.key);
		});
	}

	registerKeyEvent(
		event: string | Array<string>,
		outputValue: T,
		type: KeyEventType = KeyEventType.AtLeastOnce
	): void {
		const addKey = (v: string): void => {
			this.keyMap.set(v, {
				outputValue,
				eventType: type,
				polled: false,
			});
		};
		if (typeof event === "string") {
			addKey(event);
		} else {
			event.forEach((e) => {
				addKey(e);
			});
		}
	}
	clear(): void {
		this.keyQueue.clear();
	}
	reset(): void {
		this.clear();
		this.keyMap.clear();
	}
	pollSpecificKey(key: T, callback: () => void): void {
		const v = this.keyQueue.get(key);
		if (v === null || v === undefined) {
			return;
		}
		if (!this.keysPressed.has(v.outputValue)) {
			if (v.eventType === KeyEventType.IfCaught || v.polled) {
				v.polled = false;
				this.keyQueue.delete(v.outputValue);
				return;
			}
		}
		if (!v.polled && v.eventType === KeyEventType.OnlyOnce) {
			callback();
		} else if (v.eventType !== KeyEventType.OnlyOnce) {
			callback();
		}
		v.polled = true;


	}

	poll(): Array<T> {
		const keys: Array<T> = [];

		for (const [_, event] of this.keyQueue) {
			// Check if key in queue is still being pressed
			if (!this.keysPressed.has(event.outputValue)) {
				if (event.eventType === KeyEventType.IfCaught || event.polled) {
					event.polled = false;
					this.keyQueue.delete(event.outputValue);
					continue;
				}
			}
			if (!event.polled && event.eventType === KeyEventType.OnlyOnce) {
				keys.push(event.outputValue);
			}
			if (event.eventType !== KeyEventType.OnlyOnce) {
				keys.push(event.outputValue);
			}
			event.polled = true;
		}

		return keys;
	}

	private keyDown(key: string): void {
		const mapped = this.keyMap.get(key);
		if (mapped === null || mapped === undefined) {
			return;
		}
		this.keysPressed.add(mapped.outputValue);
		if (!this.keyQueue.has(mapped.outputValue)) {
			this.keyQueue.set(mapped.outputValue, mapped);
		}
	}
	private keyUp(key: string): void {
		const mapped = this.keyMap.get(key);
		if (mapped === null || mapped === undefined) {
			return;
		}
		this.keysPressed.delete(mapped.outputValue);
	}
}
